import enum
import time
import struct

from collections import namedtuple
from datetime import datetime, timezone

UNIX_EPOCH = datetime(1970, 1, 1, tzinfo=timezone.utc)
DTN_EPOCH = datetime(2000, 1, 1, tzinfo=timezone.utc)
UNIX_TO_DTN_OFFSET = (DTN_EPOCH - UNIX_EPOCH).total_seconds()
assert UNIX_TO_DTN_OFFSET == 946684800


def unix2dtn(unix_timestamp):
    """Converts a given Unix timestamp into a DTN timestamp

    Its inversion is :func:`dtn2unix`.

    Args:
        unix_timestamp: Unix timestamp
    Returns:
        numeric: DTN timestamp
    """
    return unix_timestamp - UNIX_TO_DTN_OFFSET


class RouterCommand(enum.IntEnum):
    """uD3TN Command Constants"""
    ADD = 1
    UPDATE = 2
    DELETE = 3
    QUERY = 4


Contact = namedtuple('Contact', ['start', 'end', 'bitrate'])
Contact.__doc__ = """named tuple holding uD3TN contact information

Attrs:
    start (int): DTN timestamp when the contact starts
    end (int): DTN timestamp when the contact is over
    bitrate (int): Bitrate of the contact
"""


def make_contact(start_offset, duration, bitrate):
    """Create a :class:`Contact` tuple relative to the current time

    Args:
        start_offset (int): Start point of the contact from in seconds from now
        duration (int): Duration of the contact in seconds
        bitrate (int): Bitrate of the contact
    Returns:
        Contact: contact tuple with DTN timestamps
    """
    cur_time = time.time()
    start = unix2dtn(cur_time + start_offset)

    return Contact(
        start=int(round(start)),
        end=int(round(start + duration)),
        bitrate=int(round(bitrate)),
    )


class ConfigMessage(object):
    """uD3TN configuration message that can be processes by its config agent.
    These messages are used to configure contacts in uD3TN.

    Args:
        eid (str): The endpoint identifier of a contact
        cla_address (str): The Convergency Layer Adapter (CLA) address for the
            contact's EID
        reachable_eids (List[str], optional): List of reachable EIDs via this
            contact
        contacts (List[Contact], optional): List of contacts with the node
        type (RouterCommand, optional): Type of the configuration message (add,
            remove, ...)
    """

    def __init__(self, eid, cla_address, reachable_eids=None, contacts=None,
                 type=RouterCommand.ADD):
        self.eid = eid
        self.cla_address = cla_address
        self.reachable_eids = reachable_eids or []
        self.contacts = contacts or []
        self.type = type

    def __repr__(self):
        return "<ConfigMessage {!r} {} reachable={} contacts={}>".format(
            self.eid, self.cla_address, self.reachable_eids, self.contacts
        )

    def __str__(self):
        # missing escaping has to be addresses in uD3TN
        for part in [self.eid, self.cla_address] + self.reachable_eids:
            assert "(" not in part
            assert ")" not in part

        if self.reachable_eids:
            eid_list = "[" + ",".join(
                "(" + eid + ")" for eid in self.reachable_eids
            ) + "]"
        else:
            eid_list = ""

        if self.contacts:
            contact_list = (
                    "[" +
                    ",".join(
                        "{{{},{},{}}}".format(start, end, bitrate)
                        for start, end, bitrate in self.contacts
                    ) +
                    "]"
                )
        else:
            contact_list = ""

        return "{}({}):({}):{}:{};".format(
            self.type,
            self.eid,
            self.cla_address,
            eid_list,
            contact_list,
        )

    def __bytes__(self):
        return str(self).encode('ascii')


class ManagementCommand(enum.IntEnum):
    """uD3TN Management Command Constants"""
    SET_TIME = 0


def serialize_set_time_cmd(unix_timestamp):
    dtn_timestamp = int(unix2dtn(unix_timestamp))
    binary = [
        # Header
        struct.pack('B', int(ManagementCommand.SET_TIME)),
        struct.pack('!Q', dtn_timestamp),
    ]
    return b"".join(binary)



#!/usr/bin/env python3
# encoding: utf-8

import logging
import socket
import uuid

from .aap_message import AAPMessage, AAPMessageType, InsufficientAAPDataError


logger = logging.getLogger(__name__)


class AAPClient():
    """A context manager class for connecting to the AAP socket of a uD3TN
    instance.

    Args:
        socket: A `socket.socket` object
        address: The address of the remote socket to be used when calling
            `socket.connect()`
    """

    def __init__(self, socket, address):
        self.socket = socket
        self.address = address
        self.node_eid = None
        self.agent_id = None

    def connect(self):
        """Establish a socket connection to a uD3TN instance and return the
        received welcome message.
        """
        self.socket.connect(self.address)
        logger.info("Connected to uD3TN, awaiting WELCOME message...")
        return self._welcome()

    def disconnect(self):
        """Shutdown and close the socket."""
        logger.info("Terminating connection...")
        self.socket.shutdown(socket.SHUT_RDWR)
        self.socket.close()

    def __enter__(self):
        """Return `self` upon calling `self.connect()` to establish the socket
        connection.
        """
        self.connect()
        return self

    def __exit__(self, type, value, traceback):
        """Invoke `self.disconnect()` on any raised runtime exception."""
        self.disconnect()

    def _welcome(self):
        """Receive the AAP welcome message and store the node EID of the uD3TN
        instance in `self.node_eid`.
        """
        msg_welcome = self.receive()
        assert msg_welcome.msg_type == AAPMessageType.WELCOME
        logger.info(f"WELCOME message received! ~ EID = {msg_welcome.eid}")
        self.node_eid = msg_welcome.eid
        return msg_welcome

    @property
    def eid(self):
        """Return the own EID."""
        return f"{self.node_eid}/{self.agent_id}"

    def register(self, agent_id=None):
        """Attempt to register the specified agent identifier.

        Args:
            agent_id: The agent identifier to be registered. If None,
                uuid.uuid4() is called to generate one.
        """
        self.agent_id = agent_id or str(uuid.uuid4())
        logger.info(f"Sending REGISTER message for '{agent_id}'...")
        msg_ack = self.send(
            AAPMessage(AAPMessageType.REGISTER, self.agent_id)
        )
        assert msg_ack.msg_type == AAPMessageType.ACK
        logger.info("ACK message received!")

    def ping(self):
        """Send a PING message via AAP and returns the ACK message (e.g. for
        keepalive purposes).
        """
        msg_ack = self.send(AAPMessage(AAPMessageType.PING))
        assert msg_ack.msg_type == AAPMessageType.ACK
        return msg_ack

    def receive(self):
        """Receive and return the next `AAPMessage`."""
        buf = bytearray()
        msg = None
        while msg is None:
            data = self.socket.recv(1)
            if not data:
                logger.info("Disconnected")
                return None
            buf += data
            try:
                msg = AAPMessage.parse(buf)
            except InsufficientAAPDataError:
                continue
        return msg

    def send(self, aap_msg):
        """Serialize and send the provided `AAPMessage` to the AAP endpoint.

        Args:
            aap_msg: The `AAPMessage` to be sent.
        """
        self.socket.send(aap_msg.serialize())
        return self.receive()

    def send_bundle(self, dest_eid, bundle_data):
        """Send the provided bundle to the AAP endpoint.

        Args:
            dest_eid: The destination EID.
            bundle_data: The binary payload data to be encapsulated in a
                bundle.
        """
        logger.info(f"Sending SENDBUNDLE message to {dest_eid}")
        msg_sendconfirm = self.send(AAPMessage(
            AAPMessageType.SENDBUNDLE, dest_eid, bundle_data
        ))
        assert msg_sendconfirm.msg_type == AAPMessageType.SENDCONFIRM
        logger.info(
            f"SENDCONFIRM message received! ~ ID = {msg_sendconfirm.bundle_id}"
        )
        return msg_sendconfirm

    def send_str(self, dest_eid, bundle_data):
        """Send the provided bundle to the AAP endpoint.

        Args:
            dest_eid: The destination EID.
            bundle_data: The string message to be utf-8 encoded and
                encapsulated in a bundle.
        """
        return self.send_bundle(dest_eid, bundle_data.encode("utf-8"))


class AAPUnixClient(AAPClient):
    """A context manager class for connecting to the AAP Unix socket of a uD3TN
    instance.

    Args:
        address: The address (PATH) of the remote socket to be used when
            calling `socket.connect()`
    """

    def __init__(self, address='/tmp/ud3tn.socket'):
        super().__init__(
            socket=socket.socket(socket.AF_UNIX, socket.SOCK_STREAM),
            address=address,
        )


class AAPTCPClient(AAPClient):
    """A context manager class for connecting to the AAP TCP socket of a uD3TN
    instance.

    Args:
        address: The address tupel (HOST, PORT) of the remote socket to be used
            when calling `socket.connect()`
    """

    def __init__(self, address=('localhost', 4242)):
        super().__init__(
            socket=socket.socket(socket.AF_INET, socket.SOCK_STREAM),
            address=address,
        )



from .aap_client import AAPClient, AAPTCPClient, AAPUnixClient
from .aap_message import AAPMessage, AAPMessageType, InsufficientAAPDataError

__all__ = [
    'AAPClient',
    'AAPTCPClient',
    'AAPUnixClient',
    'AAPMessage',
    'AAPMessageType',
    'InsufficientAAPDataError'
]



"""μD3TN Application Agent Protocol (AAP) Python implementation.

Usage:
    Use the AAPMessage class to build, serialize, and parse AAP messages.
    For a description of the protocol, see
    https://gitlab.com/d3tn/ud3tn/-/blob/master/doc/ud3tn_aap.md.
"""

import struct
import enum


class InsufficientAAPDataError(Exception):
    """The binary AAP message appears to be valid but is missing some bytes.

    Args:
        bytes_needed: The total amount of bytes required to perform the next
                      parsing step.
    """

    def __init__(self, bytes_needed):
        self.bytes_needed = bytes_needed


class AAPMessageType(enum.IntEnum):
    """AAP message type codes."""
    ACK = 0x0
    NACK = 0x1
    REGISTER = 0x2
    SENDBUNDLE = 0x3
    RECVBUNDLE = 0x4
    SENDCONFIRM = 0x5
    CANCELBUNDLE = 0x6
    WELCOME = 0x7
    PING = 0x8


class AAPMessage:
    """An AAP message representation supporting parsing and serialization."""
    def __init__(self, msg_type, eid=None, payload=None, bundle_id=None):
        if (msg_type < AAPMessageType.ACK or msg_type > AAPMessageType.PING):
            raise ValueError("Invalid message type code")
        self.msg_type = msg_type
        self.eid = eid
        self.payload = payload
        self.bundle_id = bundle_id

    def serialize(self):
        """Serialize the AAP message to its on-wire representation."""
        msg = [struct.pack("B", 0x10 | (self.msg_type & 0xF))]

        # EID
        if self.msg_type in (AAPMessageType.REGISTER,
                             AAPMessageType.SENDBUNDLE,
                             AAPMessageType.RECVBUNDLE,
                             AAPMessageType.WELCOME):
            msg.append(struct.pack("!H", len(self.eid)))
            msg.append(self.eid.encode("ascii"))

        # Payload
        if self.msg_type in (AAPMessageType.SENDBUNDLE,
                             AAPMessageType.RECVBUNDLE):
            msg.append(struct.pack("!Q", len(self.payload)))
            msg.append(self.payload)

        # Bundle ID
        if self.msg_type in (AAPMessageType.SENDCONFIRM,
                             AAPMessageType.CANCELBUNDLE):
            msg.append(struct.pack("!Q", len(self.bundle_id)))

        return b"".join(msg)

    def __bytes__(self):
        return self.serialize()

    @staticmethod
    def parse(data):
        """Parse the provided AAP on-wire representation.

        An instance of AAPMessage is returned on success.
        If there are not enough bytes provided, an InsufficientAAPDataError is
        raised, allowing the caller to wait for more data.
        If invalid data is provided, a ValueError is raised.
        """
        if len(data) == 0:
            raise InsufficientAAPDataError(1)

        version = (data[0] >> 4) & 0xF
        if version != 1:
            raise ValueError("Invalid AAP version: " + str(version))

        msg_type = data[0] & 0xF

        eid = payload = bundle_id = None
        index = 1

        # EID
        if msg_type in (AAPMessageType.REGISTER,
                        AAPMessageType.SENDBUNDLE,
                        AAPMessageType.RECVBUNDLE,
                        AAPMessageType.WELCOME):
            if len(data) - index < 2:
                raise InsufficientAAPDataError(index + 2)
            eid_length, = struct.unpack("!H", data[index:(index + 2)])
            index += 2
            if len(data) - index < eid_length:
                raise InsufficientAAPDataError(index + eid_length)
            eid = data[index:(index + eid_length)].decode("ascii")
            index += eid_length

        # Payload
        if msg_type in (AAPMessageType.SENDBUNDLE,
                        AAPMessageType.RECVBUNDLE):
            if len(data) - index < 8:
                raise InsufficientAAPDataError(index + 8)
            payload_length, = struct.unpack("!Q", data[index:(index + 8)])
            index += 8
            if len(data) - index < payload_length:
                raise InsufficientAAPDataError(index + payload_length)
            payload = data[index:(index + payload_length)]
            index += payload_length

        # Bundle ID
        if msg_type in (AAPMessageType.SENDCONFIRM,
                        AAPMessageType.CANCELBUNDLE):
            if len(data) - index < 8:
                raise InsufficientAAPDataError(index + 8)
            bundle_id, = struct.unpack("!Q", data[index:(index + 8)])
            index += 8

        return AAPMessage(msg_type, eid, payload, bundle_id)



#!/usr/bin/env python

import setuptools

with open('README.md', 'r') as file:
    long_description = file.read()

setuptools.setup(
    name='ud3tn-utils',
    version='0.9.0',
    author='D3TN GmbH',
    author_email='contact@d3tn.com',
    description='μD3TN Utilities',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://gitlab.com/d3tn/ud3tn',
    packages=['ud3tn_utils', 'ud3tn_utils.aap'],
    python_requires='>=3.6',
)
